;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "parsebib" "6.7"
  "A library for parsing bib files."
  '((emacs "25.1"))
  :url "https://github.com/joostkremers/parsebib"
  :commit "7bfde4e4679413424a9a9af099203d5c23e32cd2"
  :revdesc "7bfde4e46794"
  :keywords '("text" "bibtex")
  :authors '(("Joost Kremers" . "joostkremers@fastmail.fm"))
  :maintainers '(("Joost Kremers" . "joostkremers@fastmail.fm")))
